# -*- coding: utf-8 -*-
"""Runtime type checking with support for lazy import paths."""

from itertools import chain
from typing import TYPE_CHECKING, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import string_types

from ..utils.lazy_import import decorate_path, import_path

if TYPE_CHECKING:
    from typing import Any, Iterable, Optional, Tuple, Type, Union

    _LazyType = Union[Type, str]
    LazyTypes = Union[Optional[_LazyType], Iterable[Optional[_LazyType]]]
    LazyTypesTuple = Tuple[_LazyType, ...]
else:
    LazyTypes = None
    LazyTypesTuple = None

__all__ = [
    "LazyTypes",
    "LazyTypesTuple",
    "format_types",
    "get_type_names",
    "flatten_types",
    "import_types",
    "is_instance",
    "is_subclass",
    "assert_is_instance",
    "assert_is_subclass",
    "assert_is_callable",
]


def format_types(types, module=None):
    # type: (LazyTypes, Optional[str]) -> LazyTypesTuple
    """
    Check and format types by adding a module path.

    .. code:: python

        >>> from objetto.utils.type_checking import format_types, get_type_names

        >>> get_type_names(format_types((int, "chain"), module="itertools"))
        ('int', 'chain')
        >>> get_type_names(format_types(float))
        ('float',)
        >>> get_type_names(format_types("itertools|chain"))
        ('chain',)

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param module: Module to prefix lazy paths without one.
    :type module: str or None

    :return: Formatted types.
    :rtype: tuple[str or type or None]

    :raises ValueError: Invalid type path/no module provided.
    :raises TypeError: Did not provide valid types.
    """
    if types is None:
        return (type(None),)
    if isinstance(types, type):
        return (types,)
    elif isinstance(types, string_types):
        return (decorate_path(types, module=module),)
    elif isinstance(types, collections_abc.Iterable):
        return tuple(chain.from_iterable(format_types(t, module=module) for t in types))
    else:
        error = "'{}' instance is not a valid type".format(type(types).__name__)
        raise TypeError(error)


def get_type_names(types):
    # type: (LazyTypes) -> Tuple[str, ...]
    """
    Get type names without importing lazy paths.

    .. code:: python

        >>> from objetto.utils.type_checking import get_type_names

        >>> get_type_names((int, "itertools|chain"))
        ('int', 'chain')

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :return: Type names.
    :rtype: tuple[str]
    """
    return tuple(
        t.split("|")[-1].split(".")[-1]
        if isinstance(t, string_types)
        else (t.__name__ if isinstance(t, type) else type(t).__name__)
        for t in flatten_types(types)
    )


def flatten_types(types):
    # type: (LazyTypes) -> Tuple[_LazyType, ...]
    """
    Flatten types into a tuple.

    .. code:: python

        >>> from objetto.utils.type_checking import flatten_types, get_type_names

        >>> get_type_names(flatten_types((int, ("itertools|chain", float, (str,)))))
        ('int', 'chain', 'float', 'str')
        >>> get_type_names(flatten_types(int))
        ('int',)

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :return: Flattened types.
    :rtype: tuple[str or type]

    :raises TypeError: Invalid types.
    """
    if types is None:
        return (type(None),)
    elif isinstance(types, (string_types, type)):
        return (types,)
    elif isinstance(types, collections_abc.Iterable):
        return tuple(chain.from_iterable(flatten_types(t) for t in types))
    else:
        raise TypeError(type(types).__name__)


def import_types(types):
    # type: (LazyTypes) -> Tuple[Type, ...]
    """
    Import types from lazy import paths.

    .. code:: python

        >>> from objetto.utils.type_checking import import_types, get_type_names

        >>> get_type_names(import_types("itertools|chain"))
        ('chain',)
        >>> get_type_names(import_types(("itertools|chain", "itertools|compress")))
        ('chain', 'compress')
        >>> get_type_names(import_types(("itertools|chain", int)))
        ('chain', 'int')

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :return: Imported types.
    :rtype: tuple[type]
    """
    return tuple(
        cast(type, import_path(t)) if isinstance(t, string_types) else t
        for t in flatten_types(types)
    )


def is_instance(obj, types, subtypes=True):
    # type: (Any, LazyTypes, bool) -> bool
    """
    Get whether object is an instance of any of the provided types.

    .. code:: python

        >>> from itertools import chain
        >>> from objetto.utils.type_checking import is_instance

        >>> class SubChain(chain):
        ...     pass
        ...
        >>> is_instance(3, int)
        True
        >>> is_instance(3, (chain, int))
        True
        >>> is_instance(3, ())
        False
        >>> is_instance(SubChain(), "itertools|chain")
        True
        >>> is_instance(chain(), "itertools|chain", subtypes=False)
        True
        >>> is_instance(SubChain(), "itertools|chain", subtypes=False)
        False

    :param obj: Object.
    :type obj: object

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :return: True if it is an instance.
    :rtype: bool
    """
    imported_types = import_types(types)
    if subtypes:
        return isinstance(obj, imported_types)
    else:
        return type(obj) in imported_types


def is_subclass(cls, types, subtypes=True):
    # type: (type, LazyTypes, bool) -> bool
    """
    Get whether class is a subclass of any of the provided types.

    .. code:: python

        >>> from itertools import chain
        >>> from objetto.utils.type_checking import is_subclass

        >>> class SubChain(chain):
        ...     pass
        ...
        >>> is_subclass(int, int)
        True
        >>> is_subclass(int, (chain, int))
        True
        >>> is_subclass(int, ())
        False
        >>> is_subclass(SubChain, "itertools|chain")
        True
        >>> is_subclass(chain, "itertools|chain", subtypes=False)
        True
        >>> is_subclass(SubChain, "itertools|chain", subtypes=False)
        False

    :param cls: Class.
    :type cls: type

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :return: True if it is a subclass.
    :rtype: bool

    :raises TypeError: Did not provide a class.
    """
    if not isinstance(cls, type):
        error = "is_subclass() arg 1 must be a class"
        raise TypeError(error)
    imported_types = import_types(types)
    if subtypes:
        return issubclass(cls, import_types(types))
    else:
        return cls in imported_types


def assert_is_instance(obj, types, subtypes=True):
    # type: (Any, LazyTypes, bool) -> None
    """
    Assert object is an instance of any of the provided types.

    .. code:: python

        >>> from itertools import chain
        >>> from objetto.utils.type_checking import assert_is_instance

        >>> class SubChain(chain):
        ...     pass
        ...
        >>> assert_is_instance(3, int)
        >>> assert_is_instance(3, (chain, int))
        >>> assert_is_instance(3, ())
        Traceback (most recent call last):
        ValueError: no types were provided to perform assertion
        >>> assert_is_instance(3, "itertools|chain")
        Traceback (most recent call last):
        TypeError: got 'int' object, expected instance of 'chain' or any of its \
subclasses
        >>> assert_is_instance(chain(), "itertools|chain", subtypes=False)
        >>> assert_is_instance(SubChain(), "itertools|chain", subtypes=False)
        Traceback (most recent call last):
        TypeError: got 'SubChain' object, expected instance of 'chain' (instances of \
subclasses are not accepted)

    :param obj: Object.
    :type obj: object

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :raises ValueError: No types were provided.
    :raises TypeError: Object is not an instance of provided types.
    """
    if not is_instance(obj, types, subtypes=subtypes):
        types = flatten_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)
        error = "got '{}' object, expected instance of {}{}".format(
            type(obj).__name__,
            ", ".join("'{}'".format(name) for name in get_type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (instances of subclasses are not accepted)",
        )
        raise TypeError(error)


def assert_is_subclass(cls, types, subtypes=True):
    # type: (type, LazyTypes, bool) -> None
    """
    Assert a class is a subclass of any of the provided types.

    .. code:: python

        >>> from itertools import chain
        >>> from objetto.utils.type_checking import assert_is_subclass

        >>> class SubChain(chain):
        ...     pass
        ...
        >>> assert_is_subclass(int, int)
        >>> assert_is_subclass(int, (chain, int))
        >>> assert_is_subclass(int, ())
        Traceback (most recent call last):
        ValueError: no types were provided to perform assertion
        >>> assert_is_subclass(int, "itertools|chain")
        Traceback (most recent call last):
        TypeError: got 'int', expected class 'chain' or any of its subclasses
        >>> assert_is_subclass(chain, "itertools|chain", subtypes=False)
        >>> assert_is_subclass(SubChain, "itertools|chain", subtypes=False)
        Traceback (most recent call last):
        TypeError: got 'SubChain', expected class 'chain' (subclasses are not accepted)

    :param cls: Class.
    :type cls: type

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :raises ValueError: No types were provided.
    :raises TypeError: Class is not a subclass of provided types.
    """
    if not is_instance(cls, type):
        types = flatten_types(types)
        error = "got instance of '{}', expected {}{}{}".format(
            type(cls).__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join("'{}'".format(name) for name in get_type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (subclasses are not accepted)",
        )
        raise TypeError(error)

    if not is_subclass(cls, types, subtypes=subtypes):
        types = flatten_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)

        error = "got '{}', expected {}{}{}".format(
            cls.__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join("'{}'".format(name) for name in get_type_names(types)),
            " or any of {} subclasses".format("their" if len(types) > 1 else "its")
            if subtypes
            else " (subclasses are not accepted)",
        )
        raise TypeError(error)


def assert_is_callable(value):
    # type: (Any) -> None
    """
    Assert a value is callable.

    .. code:: python

        >>> from objetto.utils.type_checking import assert_is_subclass

        >>> assert_is_callable(int)
        >>> assert_is_callable(lambda: None)
        >>> assert_is_callable(3)
        Traceback (most recent call last):
        TypeError: got non-callable 'int' object, expected a callable

    :param value: Value.

    :raises TypeError: Value is not a match.
    """
    if not callable(value):
        error = "got non-callable '{}' object, expected a callable".format(
            type(value).__name__,
        )
        raise TypeError(error)



# -*- coding: utf-8 -*-
"""Base exceptions."""

__all__ = ["BaseObjettoException"]


class BaseObjettoException(Exception):
    """
    Base `Objetto` exception.

    Inherits from:
      - :class:`Exception`

    Inherited By:
      - :class:`objetto.exceptions.ActionObserversFailedError`
      - :class:`objetto.exceptions.RejectChangeException`
      - :class:`objetto.exceptions.HistoryError`
    """



# -*- coding: utf-8 -*-
"""Observer mixin class."""

from ._observers import ActionObserver, ActionObserverExceptionData, ActionObserverToken

__all__ = ["ActionObserver", "ActionObserverToken", "ActionObserverExceptionData"]



# -*- coding: utf-8 -*-
"""Data."""

from typing import TYPE_CHECKING, TypeVar

from ._bases import MISSING
from ._data import (
    Data,
    DataAttribute,
    DataRelationship,
    DictData,
    InteractiveData,
    InteractiveDictData,
    InteractiveListData,
    InteractiveSetData,
    ListData,
    SetData,
)
from ._structures import (
    KeyRelationship,
    UniqueDescriptor,
    make_auxiliary_cls,
    unique_descriptor,
)
from .utils.caller_module import get_caller_module
from .utils.reraise_context import ReraiseContext

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Optional, Type, Union

    from .utils.factoring import LazyFactory

__all__ = [
    "DataRelationship",
    "KeyRelationship",
    "UniqueDescriptor",
    "Data",
    "InteractiveData",
    "DictData",
    "InteractiveDictData",
    "ListData",
    "InteractiveListData",
    "SetData",
    "InteractiveSetData",
    "DataAttribute",
    "unique_descriptor",
    "data_attribute",
    "data_constant_attribute",
    "data_dict_attribute",
    "data_list_attribute",
    "data_set_attribute",
    "data_dict_cls",
    "data_list_cls",
    "data_set_cls",
]


T = TypeVar("T")  # Any type.
KT = TypeVar("KT")  # Any key type.
VT = TypeVar("VT")  # Any value type.


if TYPE_CHECKING:
    NT = Union[Type[None], None]
    InteractiveDictAttribute = DataAttribute[InteractiveDictData[KT, VT]]
    InteractiveListAttribute = DataAttribute[InteractiveListData[T]]
    InteractiveSetAttribute = DataAttribute[InteractiveSetData[T]]
    DictAttribute = DataAttribute[DictData[KT, VT]]
    ListAttribute = DataAttribute[ListData[T]]
    SetAttribute = DataAttribute[SetData[T]]


def data_attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=True,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
):
    # type: (...) -> DataAttribute[T]
    """
    Make data attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :return: Data attribute.
    :rtype: objetto.data.DataAttribute

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_attribute'"):
        relationship = DataRelationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'data_attribute'"):
        attribute = DataAttribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: DataAttribute[T]

    return attribute


def data_constant_attribute(
    value,  # type: T
    serialized=False,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=False,  # type: bool
    compared=True,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
):
    # type: (...) -> DataAttribute[T]
    """
    Make constant data attribute.

    :param value: Constant value.

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :return: Constant data attribute.
    :rtype: objetto.data.DataAttribute

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_constant_attribute'"):
        relationship = DataRelationship(
            types=type(value),
            subtypes=False,
            checked=False,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'data_constant_attribute'"):
        attribute_ = DataAttribute(
            relationship=relationship,
            default=value,
            default_factory=None,
            required=False,
            changeable=False,
            deletable=False,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: DataAttribute[T]

    return attribute_


def data_dict_attribute(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    key_types=(),  # type: Union[Type[KT], NT, str, Iterable[Union[Type[KT], NT, str]]]
    key_subtypes=False,  # type: bool
    key_factory=None,  # type: LazyFactory
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[InteractiveDictAttribute[KT, VT], DictAttribute[KT, VT]]
    """
    Make dictionary data attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param key_types: Key types.
    :type key_types: str or type or None or tuple[str or type or None]

    :param key_subtypes: Whether to accept subtypes for the keys.
    :type key_subtypes: bool

    :param key_factory: Key factory.
    :type key_factory: str or collections.abc.Callable or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: Dictionary data attribute.
    :rtype: objetto.data.DataAttribute[objetto.data.InteractiveDictData or \
objetto.data.DictData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make dictionary class.
    with ReraiseContext((TypeError, ValueError), "defining 'data_dict_attribute'"):
        dict_type = data_dict_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            compared=compared,
            key_types=key_types,
            key_subtypes=key_subtypes,
            key_factory=key_factory,
            qual_name=qual_name,
            unique=unique,
            interactive=interactive,
        )  # type: Union[Type[DictData[KT, VT]], Type[InteractiveDictData[KT, VT]]]

    # Factory that forces the dictionary type.
    def dict_factory(initial=()):
        if type(initial) is dict_type:
            return initial
        else:
            return dict_type(initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = {}

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_dict_attribute'"):
        relationship = DataRelationship(
            types=dict_type,
            subtypes=False,
            checked=checked,
            module=module,
            factory=dict_factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'data_dict_attribute'"):
        attribute_ = DataAttribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: DataAttribute[InteractiveDictData[KT, VT]]

    return attribute_


def data_list_attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[InteractiveListAttribute[T], ListAttribute[T]]
    """
    Make list data attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: List data attribute.
    :rtype: objetto.data.DataAttribute[objetto.data.InteractiveListData or \
objetto.data.ListData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make list class.
    with ReraiseContext((TypeError, ValueError), "defining 'data_list_attribute'"):
        list_type = data_list_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            compared=compared,
            qual_name=qual_name,
            unique=unique,
            interactive=interactive,
        )  # type: Union[Type[ListData[T]], Type[InteractiveListData[T]]]

    # Factory that forces the list type.
    def list_factory(initial=()):
        if type(initial) is list_type:
            return initial
        else:
            return list_type(initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = {}

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_list_attribute'"):
        relationship = DataRelationship(
            types=list_type,
            subtypes=False,
            checked=checked,
            module=module,
            factory=list_factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'data_list_attribute'"):
        attribute_ = DataAttribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: DataAttribute[InteractiveListData[T]]

    return attribute_


def data_set_attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[InteractiveSetAttribute[T], SetAttribute[T]]
    """
    Make set data attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: Set data attribute.
    :rtype: objetto.data.DataAttribute[objetto.data.InteractiveSetData or \
objetto.data.SetData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make set class.
    with ReraiseContext((TypeError, ValueError), "defining 'data_set_attribute'"):
        set_type = data_set_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            compared=compared,
            qual_name=qual_name,
            unique=unique,
            interactive=interactive,
        )  # type: Union[Type[SetData[T]], Type[InteractiveSetData[T]]]

    # Factory that forces the set type.
    def set_factory(initial=()):
        if type(initial) is set_type:
            return initial
        else:
            return set_type(initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = {}

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_set_attribute'"):
        relationship = DataRelationship(
            types=set_type,
            subtypes=False,
            checked=checked,
            module=module,
            factory=set_factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'data_set_attribute'"):
        attribute_ = DataAttribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: DataAttribute[InteractiveSetData[T]]

    return attribute_


def data_dict_cls(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    key_types=(),  # type: Union[Type[KT], NT, str, Iterable[Union[Type[KT], NT, str]]]
    key_subtypes=False,  # type: bool
    key_factory=None,  # type: LazyFactory
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[Type[InteractiveDictData[KT, VT]], Type[DictData[KT, VT]]]
    """
    Make auxiliary dictionary data class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param key_types: Key types.
    :type key_types: str or type or None or tuple[str or type or None]

    :param key_subtypes: Whether to accept subtypes for the keys.
    :type key_subtypes: bool

    :param key_factory: Key factory.
    :type key_factory: str or collections.abc.Callable or None

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: Dictionary data class.
    :type: type[objetto.data.DictData or objetto.data.InteractiveDictData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_dict_cls'"):
        relationship = DataRelationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Key relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_dict_cls'"):
        key_relationship = KeyRelationship(
            types=key_types,
            subtypes=key_subtypes,
            checked=checked,
            module=module,
            factory=key_factory,
        )
        dct = {"_key_relationship": key_relationship}

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
        dct=dct,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'data_dict_cls'"):
        if interactive:
            interactive_base = (
                InteractiveDictData
            )  # type: Type[InteractiveDictData[KT, VT]]
            return make_auxiliary_cls(interactive_base, **cls_kwargs)
        else:
            base = DictData  # type: Type[DictData[KT, VT]]
            return make_auxiliary_cls(base, **cls_kwargs)


def data_list_cls(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[Type[InteractiveListData[T]], Type[ListData[T]]]
    """
    Make auxiliary list data class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: List data class.
    :type: type[objetto.data.ListData or objetto.data.InteractiveListData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_list_cls'"):
        relationship = DataRelationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'data_list_cls'"):
        if interactive:
            interactive_base = InteractiveListData  # type: Type[InteractiveListData[T]]
            return make_auxiliary_cls(interactive_base, **cls_kwargs)
        else:
            base = ListData  # type: Type[ListData[T]]
            return make_auxiliary_cls(base, **cls_kwargs)


def data_set_cls(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    interactive=True,  # type: bool
):
    # type: (...) -> Union[Type[InteractiveSetData[T]], Type[SetData[T]]]
    """
    Make auxiliary set data class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param interactive: Whether generated class should be interactive.
    :type interactive: bool

    :return: Set data class.
    :type: type[objetto.data.SetData or objetto.data.InteractiveSetData]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'data_set_cls'"):
        relationship = DataRelationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            compared=compared,
        )

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'data_set_cls'"):
        if interactive:
            interactive_base = InteractiveSetData  # type: Type[InteractiveSetData[T]]
            return make_auxiliary_cls(interactive_base, **cls_kwargs)
        else:
            base = SetData  # type: Type[SetData[T]]
            return make_auxiliary_cls(base, **cls_kwargs)



# -*- coding: utf-8 -*-
"""State-carrying structures."""

from abc import abstractmethod
from inspect import getmro
from re import sub as re_sub
from typing import TYPE_CHECKING, TypeVar, cast, overload
from weakref import WeakKeyDictionary

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, string_types, with_metaclass

from .._bases import (
    FINAL_METHOD_TAG,
    Base,
    BaseHashable,
    BaseInteractiveCollection,
    BaseMeta,
    BaseMutableCollection,
    BaseProtectedCollection,
    final,
    make_base_cls,
)
from .._exceptions import BaseObjettoException
from .._states import BaseState
from ..utils.custom_repr import custom_mapping_repr
from ..utils.factoring import format_factory, import_factory, run_factory
from ..utils.lazy_import import get_path, import_path
from ..utils.recursive_repr import recursive_repr
from ..utils.reraise_context import ReraiseContext
from ..utils.type_checking import (
    assert_is_instance,
    format_types,
    get_type_names,
    import_types,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        List,
        Mapping,
        MutableMapping,
        Optional,
        Tuple,
        Type,
        Union,
    )

    from ..utils.factoring import LazyFactory
    from ..utils.type_checking import LazyTypes, LazyTypesTuple

__all__ = [
    "make_auxiliary_cls",
    "SerializationError",
    "BaseRelationship",
    "UniqueDescriptor",
    "unique_descriptor",
    "BaseStructureMeta",
    "BaseStructure",
    "BaseInteractiveStructure",
    "BaseMutableStructure",
    "BaseAuxiliaryStructureMeta",
    "BaseAuxiliaryStructure",
    "BaseInteractiveAuxiliaryStructure",
    "BaseMutableAuxiliaryStructure",
]


_SERIALIZED_CLASS_KEY = "__class__"
_ESCAPED_SERIALIZED_CLASS_KEY = "\\__class__"
_SERIALIZED_VALUE_KEY = "__state__"


T = TypeVar("T")  # Any type.


def _escape_serialized_class(dct):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Escape serialized '__class__' key.

    :param dct: Serialized dictionary.
    :return: Escaped serialized dictionary.
    """
    if _SERIALIZED_CLASS_KEY in dct:
        dct = dct.copy()
        dct[_ESCAPED_SERIALIZED_CLASS_KEY] = dct.pop(_SERIALIZED_CLASS_KEY)
    return dct


def _unescape_serialized_class(dct):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Unescape serialized '__class__' key.

    :param dct: Serialized dictionary.
    :return: Unescaped serialized dictionary.
    """
    if _ESCAPED_SERIALIZED_CLASS_KEY in dct:
        dct = dct.copy()
        dct[_SERIALIZED_CLASS_KEY] = dct.pop(_ESCAPED_SERIALIZED_CLASS_KEY)
    return dct


def _capitalize_first(string):
    # type: (str) -> str
    """
    Capitalize first letter of string, without touching the rest of it.

    :param string: String.
    :return: Capitalized string.
    """
    return string[:1].upper() + string[1:]


# noinspection PyTypeChecker
_A = TypeVar("_A", bound="Type[BaseAuxiliaryStructure]")


def make_auxiliary_cls(
    base,  # type: _A
    relationship,  # type: BaseRelationship
    qual_name=None,  # type: Optional[str]
    module=None,  # type: Optional[str]
    unique_descriptor_name=None,  # type: Optional[str]
    dct=None,  # type: Optional[Mapping[str, Any]]
):
    # type: (...) -> _A
    """
    Make an auxiliary container subclass on the fly.
    :param base: Base auxiliary container class.
    :param relationship: Relationship.
    :param qual_name: Qualified name.
    :param module: Module.
    :param unique_descriptor_name: Attribute name for unique descriptor.
    :param dct: Members dictionary.
    :return: Generated auxiliary container subclass.
    :raises TypeError: Invalid 'qual_name' parameter type.
    :raises TypeError: Invalid 'module' parameter type.
    """

    # 'qual_name'
    with ReraiseContext(TypeError, "'qual_name' parameter"):
        assert_is_instance(qual_name, string_types + (None,))

    # 'module'
    with ReraiseContext(TypeError, "'module' parameter"):
        assert_is_instance(module, string_types + (None,))
    module = module or None

    # Generate default name based on relationship types.
    if qual_name is None:
        type_names = get_type_names(
            tuple(t for t in relationship.types if t not in (type(None), None))
        )
        if not type_names:
            qual_name = base.__name__
        else:
            base_name = base.__name__
            prefix = "".join(
                _capitalize_first(re_sub(r"[^A-Za-z]+", "", tn)) for tn in type_names
            )
            qual_name = "{}{}".format(
                prefix if prefix != base_name else "",
                base_name,
            )

    # Copy dct and add relationship to it.
    dct_copy = dict(dct or {})
    dct_copy["_relationship"] = relationship

    # Add unique descriptor.
    if unique_descriptor_name:
        dct_copy[unique_descriptor_name] = UniqueDescriptor()

    return cast(
        "_A",
        make_base_cls(
            base=base,
            qual_name=qual_name,
            module=module,
            dct=dct_copy,
        ),
    )


class SerializationError(BaseObjettoException):
    """
    Failed to serialize/deserialize structure.

    Inherits from:
      - :class:`objetto.bases.BaseObjettoException`
    """


class BaseRelationship(BaseHashable):
    """
    Relationship between a structure and its values.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`

    Inherited by:
      - :class:`objetto.data.DataRelationship`
      - :class:`objetto.objects.Relationship`

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    __slots__ = (
        "__hash",
        "__types",
        "__subtypes",
        "__checked",
        "__module",
        "__factory",
        "__serialized",
        "__serializer",
        "__deserializer",
        "__represented",
    )

    def __init__(
        self,
        types=(),  # type: LazyTypes
        subtypes=False,  # type: bool
        checked=None,  # type: Optional[bool]
        module=None,  # type: Optional[str]
        factory=None,  # type: LazyFactory
        serialized=True,  # type: bool
        serializer=None,  # type: LazyFactory
        deserializer=None,  # type: LazyFactory
        represented=True,  # type: bool
    ):
        # type: (...) -> None

        # 'module'
        with ReraiseContext(TypeError, "'module' parameter"):
            assert_is_instance(module, string_types + (None,))
        module = module or None

        # 'types' and 'checked'
        with ReraiseContext((ValueError, TypeError), "'types' parameter"):
            types = format_types(types, module=module)
        if not types:
            if checked:
                error = "did not provide any 'types' but 'checked' is True"
                raise ValueError(error)
            if checked is None:
                checked = False
        else:
            if checked is None:
                checked = True
        checked = bool(checked)

        # 'factory'
        with ReraiseContext((ValueError, TypeError), "'factory' parameter"):
            factory = format_factory(factory, module=module)

        # 'serialized', 'serializer', and 'deserializer'
        if not serialized:
            if serializer is not None:
                error = "provided 'serializer' but 'serialized' is False"
                raise ValueError(error)
            if deserializer is not None:
                error = "provided 'deserializer' but 'serialized' is False"
                raise ValueError(error)
        else:
            if serializer is not None:
                with ReraiseContext((ValueError, TypeError), "'serializer' parameter"):
                    serializer = format_factory(serializer, module=module)
            if deserializer is not None:
                with ReraiseContext(
                    (ValueError, TypeError), "'deserializer' parameter"
                ):
                    deserializer = format_factory(deserializer, module=module)

        self.__hash = None  # type: Optional[int]
        self.__types = types
        self.__subtypes = bool(subtypes)
        self.__checked = checked
        self.__module = module
        self.__factory = factory
        self.__serialized = bool(serialized)
        self.__serializer = serializer
        self.__deserializer = deserializer
        self.__represented = bool(represented)

    @final
    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        if self.__hash is None:
            self.__hash = hash(frozenset(iteritems(self.to_dict())))
        return self.__hash

    @final
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare with another object for equality.

        :param other: Another object.

        :return: True if considered equal.
        :rtype: bool
        """
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        assert isinstance(other, BaseRelationship)
        return self.to_dict() == other.to_dict()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            sorting=True,
            key_repr=str,
        )

    def __str__(self):
        # type: () -> str
        """
        Get string representation.

        :return: String representation.
        :rtype: str
        """
        return self.__repr__()

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {
            "types": frozenset(import_types(self.types)),
            "subtypes": self.subtypes,
            "checked": self.checked,
            "module": self.module,
            "factory": import_factory(self.factory),
            "serialized": self.serialized,
            "serializer": import_factory(self.serializer),
            "deserializer": import_factory(self.deserializer),
            "represented": self.represented,
        }

    @final
    def get_single_exact_type(self, types=(type,)):
        # type: (LazyTypes) -> Optional[Type]
        """
        Get single exact type from available types if possible.

        :param types: Base types.
        :type types: str or type or None or tuple[str or type or None]

        :return: Single exact type that is a subclass of one of provided base types.
        :rtype: type or None
        """
        if not types or self.subtypes:
            return None
        filtered_types = set(
            typ
            for typ in import_types(self.types)
            if issubclass(typ, import_types(types))
        )
        if len(filtered_types) == 1:
            return next(iter(filtered_types))
        else:
            return None

    @final
    def fabricate_value(self, value, factory=True, **kwargs):
        # type: (Any, bool, Any) -> Any
        """
        Perform type check and run value through factory.

        :param value: Value.

        :param factory: Whether to run value through factory.
        :type factory: bool

        :param kwargs: Keyword arguments to be passed to the factory.

        :return: Fabricated value.
        """
        if factory and self.factory is not None:
            value = run_factory(self.factory, args=(value,), kwargs=kwargs)
        if self.types and self.checked:
            assert_is_instance(value, self.types, subtypes=self.subtypes)
        return value

    @property
    @final
    def types(self):
        # type: () -> LazyTypesTuple
        """
        Types.

        :rtype: tuple[str or type]
        """
        return self.__types

    @property
    @final
    def subtypes(self):
        # type: () -> bool
        """
        Whether to accept subtypes.

        :rtype: bool
        """
        return self.__subtypes

    @property
    @final
    def checked(self):
        # type: () -> bool
        """
        Whether to perform runtime type check.

        :rtype: bool
        """
        return self.__checked

    @property
    @final
    def module(self):
        # type: () -> Optional[str]
        """
        Module path for lazy types/factories.

        :rtype: str or None
        """
        return self.__module

    @property
    @final
    def factory(self):
        # type: () -> LazyFactory
        """
        Value factory.

        :rtype: str or collections.abc.Callable or None
        """
        return self.__factory

    @property
    @final
    def serialized(self):
        # type: () -> bool
        """
        Whether should be serialized.

        :rtype: bool
        """
        return self.__serialized

    @property
    @final
    def serializer(self):
        # type: () -> LazyFactory
        """
        Custom serializer.

        :rtype: str or collections.abc.Callable or None
        """
        return self.__serializer

    @property
    @final
    def deserializer(self):
        # type: () -> LazyFactory
        """
        Custom deserializer.

        :rtype: str or collections.abc.Callable or None
        """
        return self.__deserializer

    @property
    @final
    def represented(self):
        # type: () -> bool
        """
        Whether should be represented.

        :rtype: bool
        """
        return self.__represented

    @property
    @final
    def passthrough(self):
        # type: () -> bool
        """
        Whether does not perform type checks and has no factory.

        :rtype: bool
        """
        return (not self.types or not self.checked) and self.factory is None


# noinspection PyTypeChecker
_UD = TypeVar("_UD", bound="UniqueDescriptor")


@final
class UniqueDescriptor(Base):
    """
    Descriptor to be used on :class:`objetto.bases.BaseStructure` classes.
    When used, the object ID will be the hash, and the equality method will compare by
    identity instead of by values.
    If accessed through an instance, the descriptor will return the object ID.

    Inherits from:
      - :class:`objetto.bases.Base`
    """

    __slots__ = (FINAL_METHOD_TAG,)

    def __init__(self):
        setattr(self, FINAL_METHOD_TAG, True)

    @overload
    def __get__(self, instance, owner):
        # type: (_UD, None, Type[BaseStructure]) -> _UD
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (BaseStructure, Type[BaseStructure]) -> int
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (_UD, object, type) -> _UD
        pass

    def __get__(self, instance, owner):
        """
        Get object hash when accessing from instance or this descriptor otherwise.

        :param instance: Instance.
        :type instance: objetto.bases.BaseStructure or None

        :param owner: Owner class.
        :type owner: type[objetto.bases.BaseStructure] or None

        :return: Object hash based on object ID or this descriptor.
        :rtype: int or objetto.objects.UniqueDescriptor or objetto.data.UniqueDescriptor
        """
        if instance is not None:
            cls = type(instance)
            if getattr(cls, "_unique_descriptor", None) is self:
                return hash(id(instance))
        return self


# noinspection PyAbstractClass
def unique_descriptor():
    # type: () -> UniqueDescriptor
    """
    Descriptor to be used when declaring an :class:`objetto.objects.Object` or a
    :class:`objetto.data.InteractiveData` container class.
    When used, the hash for the container will be the object ID, and the equality method
    will compare by identity instead of values.
    If accessed through an instance, the descriptor will return the unique hash based
    on the object's ID.

    .. code:: python

        >>> from objetto import Application, Object, unique_descriptor

        >>> class UniqueObject(Object):
        ...     unique_hash = unique_descriptor()
        ...
        >>> app = Application()
        >>> obj = UniqueObject(app)
        >>> obj.unique_hash == hash(id(obj))
        True

    :return: Unique descriptor.
    :rtype: objetto.objects.UniqueDescriptor or objetto.data.UniqueDescriptor
    """
    return UniqueDescriptor()


class BaseStructureMeta(BaseMeta):
    """Metaclass for `BaseStructure`."""

    __unique_descriptor_name = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseStructureMeta, Optional[str]]
    __unique_descriptor = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseStructureMeta, Optional[UniqueDescriptor]]

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(BaseStructureMeta, cls).__init__(name, bases, dct)

        # Find unique descriptors.
        unique_descriptors = {}
        for base in reversed(getmro(cls)):
            base_is_base_structure = isinstance(base, BaseStructureMeta)
            for member_name, member in iteritems(base.__dict__):

                # Found unique descriptor.
                if type(member) is UniqueDescriptor:

                    # Valid declaration.
                    if base_is_base_structure:
                        unique_descriptors[member_name] = member

                    # Invalid.
                    else:
                        error = (
                            "unique descriptor '{}' can't be declared in base '{}', "
                            "which is not a subclass of '{}'"
                        ).format(member_name, base.__name__, BaseStructure.__name__)
                        raise TypeError(error)

                # Was overridden.
                elif member_name in unique_descriptors:
                    del unique_descriptors[member_name]

        # Multiple unique descriptors.
        if len(unique_descriptors) > 1:
            error = "class '{}' has multiple unique descriptors at {}".format(
                cls.__name__, ", ".join("'{}'".format(n) for n in unique_descriptors)
            )
            raise TypeError(error)

        # Store unique descriptor.
        unique_descriptor_name = None  # type: Optional[str]
        unique_descriptor_ = None  # type: Optional[UniqueDescriptor]
        if unique_descriptors:
            unique_descriptor_name, unique_descriptor_ = next(
                iteritems(unique_descriptors)
            )
        type(cls).__unique_descriptor_name[cls] = unique_descriptor_name
        type(cls).__unique_descriptor[cls] = unique_descriptor_

    @property
    @final
    def _unique_descriptor_name(cls):
        # type: () -> Optional[str]
        """Unique descriptor name or `None`."""
        return type(cls).__unique_descriptor_name[cls]

    @property
    @final
    def _unique_descriptor(cls):
        # type: () -> Optional[UniqueDescriptor]
        """Unique descriptor or `None`."""
        return type(cls).__unique_descriptor[cls]

    @property
    @abstractmethod
    def _serializable_structure_types(cls):
        # type: () -> Tuple[Type[BaseStructure], ...]
        """Serializable structure types."""
        raise NotImplementedError()

    @property
    def _relationship_type(cls):
        # type: () -> Type[BaseRelationship]
        """Relationship type."""
        return BaseRelationship


# noinspection PyTypeChecker
_BS = TypeVar("_BS", bound="BaseStructure")


class BaseStructure(
    with_metaclass(BaseStructureMeta, BaseHashable, BaseProtectedCollection[T])
):
    """
    Base structure.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveStructure`
      - :class:`objetto.bases.BaseMutableStructure`
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseData`
      - :class:`objetto.bases.BaseObject`

    Features:
      - Is hashable.
      - Is a protected collection.
      - Has state.
      - Unique hash based on ID if unique descriptor is defined.
      - Holds values at locations.
      - Has a relationship for each location.
      - Serializes/deserializes values and itself.
    """

    __slots__ = ()

    @final
    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        cls = type(self)
        if cls._unique_descriptor:
            return hash(id(self))
        else:
            return self._hash()

    @final
    def __eq__(self, other):
        # type: (Any) -> bool
        """
        Compare with another object for equality/identity.

        :param other: Another object.

        :return: True if equal or the exact same object.
        :rtype: bool
        """
        if self is other:
            return True
        if not isinstance(other, collections_abc.Hashable):
            return self._eq(other)
        cls = type(self)
        if cls._unique_descriptor:
            return False
        elif isinstance(other, BaseStructure) and type(other)._unique_descriptor:
            return False
        else:
            return self._eq(other)

    @staticmethod
    @final
    def __deserialize_value(
        serialized,  # type: Any
        location,  # type: Any
        relationship,  # type: BaseRelationship
        serializable_structure_types,  # type: Tuple[Type[BaseStructure], ...]
        class_name,  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> Any
        """
        Deserialize value for location with built-in serializer.

        :param serialized: Serialized value.
        :param location: Location.
        :param relationship: Relationship.
        :param serializable_structure_types: Serializable structure types.
        :param kwargs: Keyword arguments to be passed to the deserializers.
        :return: Deserialized value.
        :raises TypeError: Can't deserialize value due to ambiguous types.
        """

        # Possible serialized structure.
        if type(serialized) in (dict, list):

            # Serialized in a dictionary.
            if type(serialized) is dict:

                # Serialized structure with path to its class.
                if _SERIALIZED_CLASS_KEY in serialized:
                    serialized_class = import_path(
                        serialized[_SERIALIZED_CLASS_KEY]
                    )  # type: Type[BaseStructure]
                    serialized_value = serialized[_SERIALIZED_VALUE_KEY]
                    if type(serialized_value) is dict:
                        serialized_value = _unescape_serialized_class(serialized_value)
                    return serialized_class.deserialize(serialized_value, **kwargs)

                # Unescape keys.
                serialized = _unescape_serialized_class(serialized)

            # Single, non-ambiguous structure type.
            single_structure_type = relationship.get_single_exact_type(
                serializable_structure_types
            )  # type: Optional[Type[BaseStructure]]
            if single_structure_type is not None:
                return single_structure_type.deserialize(serialized, **kwargs)

            # Complex type (dict or list).
            single_complex_type = relationship.get_single_exact_type(
                (dict, list)
            )  # type: Optional[Union[Type[Dict], Type[List]]]
            if single_complex_type is None:
                error = (
                    "can't deserialize '{}' object as a value of '{}' since "
                    "relationship{} defines none or ambiguous types"
                ).format(
                    type(serialized).__name__,
                    class_name,
                    " at location {}".format(location) if location is not None else "",
                )
                raise TypeError(error)

        # Return type-check deserialized value.
        return relationship.fabricate_value(serialized, factory=False)

    @staticmethod
    @final
    def __serialize_value(
        value,  # type: Any
        relationship,  # type: BaseRelationship
        serializable_structure_types,  # type: Tuple[Type[BaseStructure], ...]
        **kwargs  # type: Any
    ):
        # type: (...) -> Any
        """
        Serialize value for location with built-in serializer.

        :param value: Value.
        :param relationship: Relationship.
        :param serializable_structure_types: Serializable structure types.
        :param kwargs: Keyword arguments to be passed to the serializers.
        :return: Serialized value.
        """

        # Structure type.
        if isinstance(value, serializable_structure_types):
            serialized_value = value.serialize(**kwargs)

            # Escape keys.
            if type(serialized_value) is dict:
                serialized_value = _escape_serialized_class(serialized_value)

            # Ambiguous type, serialize with class path.
            single_structure_type = relationship.get_single_exact_type(
                serializable_structure_types
            )  # type: Optional[Type[BaseStructure]]
            if single_structure_type is None:
                return {
                    _SERIALIZED_CLASS_KEY: get_path(type(value)),
                    _SERIALIZED_VALUE_KEY: serialized_value,
                }

            return serialized_value

        # Escape keys.
        if type(value) is dict:
            value = _escape_serialized_class(value)

        return value

    # @abstractmethod (commented out due to a PyCharm bug)
    def _hash(self):
        # type: () -> int
        """
        **Abstract**

        Get hash.

        :return: Hash.
        :rtype: int

        :raises RuntimeError: Abstract method not implemented by subclasses.
        """
        # raise NotImplementedError
        raise RuntimeError()

    # @abstractmethod (commented out due to a PyCharm bug)
    def _eq(self, other):
        # type: (Any) -> bool
        """
        **Abstract**

        Compare with another object for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool

        :raises RuntimeError: Abstract method not implemented by subclasses.
        """
        # raise NotImplementedError
        raise RuntimeError()

    @classmethod
    @abstractmethod
    def _get_relationship(cls, location):
        # type: (Any) -> BaseRelationship
        """
        Get relationship at location.

        :param location: Location.
        :type location: collections.abc.Hashable

        :return: Relationship.
        :rtype: objetto.bases.BaseRelationship

        :raises KeyError: Invalid location.
        """
        raise NotImplementedError()

    @classmethod
    @final
    def deserialize_value(cls, serialized, location=None, **kwargs):
        # type: (Any, Any, Any) -> Any
        """
        Deserialize value for location.

        :param serialized: Serialized value.

        :param location: Location.
        :type location: collections.abc.Hashable

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized value.

        :raises objetto.exceptions.SerializationError: Can't deserialize value.
        :raises ValueError: Keyword arguments contain reserved keys.
        """

        # Get relationship.
        relationship = cls._get_relationship(location)
        if not relationship.serialized:
            error = (
                "can't deserialize '{}' object as a value of '{}' since the "
                "relationship{} does not allow for serialization/deserialization"
            ).format(
                type(serialized).__name__,
                cls.__fullname__,
                " at location {}".format(location) if location is not None else "",
            )
            raise SerializationError(error)

        # Built-in deserializer.
        deserializer = lambda: cls.__deserialize_value(
            serialized,
            location,
            relationship,
            cls._serializable_structure_types,
            cls.__fullname__,
            **kwargs
        )
        if relationship.deserializer is None:
            return deserializer()

        # Check kwargs for reserved keys.
        if "super" in kwargs:
            error = "can't pass reserved keyword argument 'super' to deserializers"
            raise ValueError(error)

        # Custom deserializer.
        kwargs = dict(kwargs)
        kwargs["super"] = deserializer
        if type(serialized) is dict:
            serialized = _unescape_serialized_class(serialized)
        value = run_factory(
            relationship.deserializer, args=(serialized,), kwargs=kwargs
        )
        return relationship.fabricate_value(value, factory=False)

    @final
    def serialize_value(self, value, location=None, **kwargs):
        # type: (Any, Any, Any) -> Any
        """
        Serialize value for location.

        :param value: Value.

        :param location: Location.
        :type location: collections.abc.Hashable

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized value.

        :raises objetto.exceptions.SerializationError: Can't serialize value.
        :raises ValueError: Keyword arguments contain reserved keys.
        """

        # Get relationship.
        cls = type(self)
        relationship = cls._get_relationship(location)
        if not relationship.serialized:
            error = (
                "can't serialize '{}' value contained in a '{}' object since the "
                "relationship{} does not allow for serialization/deserialization"
            ).format(
                type(value).__name__,
                cls.__fullname__,
                " at location {}".format(location) if location is not None else "",
            )
            raise SerializationError(error)

        # Built-in serializer
        serializer = lambda: self.__serialize_value(
            value, relationship, cls._serializable_structure_types, **kwargs
        )
        if relationship.serializer is None:
            return serializer()

        # Check kwargs for reserved keys.
        if "super" in kwargs:
            error = "can't pass reserved keyword argument 'super' to serializers"
            raise ValueError(error)

        # Custom serializer.
        kwargs = dict(kwargs)
        kwargs["super"] = serializer
        serialized_value = run_factory(
            relationship.serializer, args=(value,), kwargs=kwargs
        )
        if type(serialized_value) is dict:
            serialized_value = _escape_serialized_class(serialized_value)
        return serialized_value

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized, **kwargs):
        # type: (Type[_BS], Any, Any) -> _BS
        """
        Deserialize.

        :param serialized: Serialized.

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.bases.BaseStructure

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize(self, **kwargs):
        # type: (Any) -> Any
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _state(self):
        # type: () -> BaseState
        """
        State.

        :rtype: objetto.bases.BaseState
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseInteractiveStructure(BaseStructure[T], BaseInteractiveCollection[T]):
    """
    Base interactive structure.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`
      - :class:`objetto.bases.BaseInteractiveCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`
      - :class:`objetto.bases.BaseInteractiveData`

    Features:
      - Is an interactive collection/structure.
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseMutableStructure(BaseStructure[T], BaseMutableCollection[T]):
    """
    Base mutable structure.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`
      - :class:`objetto.bases.BaseMutableCollection`

    Inherited By:
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`
      - :class:`objetto.bases.BaseMutableObject`

    Features:
      - Is a mutable collection/structure.
    """

    __slots__ = ()


class BaseAuxiliaryStructureMeta(BaseStructureMeta):
    """Metaclass for `BaseAuxiliaryStructure`."""

    def __init__(cls, name, bases, dct):
        super(BaseAuxiliaryStructureMeta, cls).__init__(name, bases, dct)

        # Check relationship type.
        relationship = getattr(cls, "_relationship")
        relationship_type = cls._relationship_type
        assert_is_instance(relationship, relationship_type, subtypes=False)

    @property
    @abstractmethod
    def _base_auxiliary_type(cls):
        # type: () -> Type[BaseAuxiliaryStructure]
        """Base auxiliary structure type."""
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseAuxiliaryStructure(
    with_metaclass(BaseAuxiliaryStructureMeta, BaseStructure[T])
):
    """
    Structure with a single relationship for all locations.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`
      - :class:`objetto.bases.BaseDictStructure`
      - :class:`objetto.bases.BaseListStructure`
      - :class:`objetto.bases.BaseSetStructure`
      - :class:`objetto.bases.BaseAuxiliaryData`
      - :class:`objetto.bases.BaseAuxiliaryObject`
    """

    __slots__ = ()

    _relationship = BaseRelationship()
    """
    **Class Attribute**

    Relationship for all locations.

    :type: objetto.bases.BaseRelationship
    """

    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        """
        return self._state.find_with_attributes(**attributes)

    @classmethod
    @final
    def _get_relationship(cls, location=None):
        # type: (Any) -> BaseRelationship
        """
        Get relationship.

        :param location: Location.
        :type location: collections.abc.Hashable

        :return: Relationship.
        :rtype: objetto.bases.BaseRelationship
        """
        return cast("BaseRelationship", cls._relationship)


# noinspection PyAbstractClass
class BaseInteractiveAuxiliaryStructure(
    BaseAuxiliaryStructure[T],
    BaseInteractiveStructure[T],
):
    """
    Base interactive auxiliary structure.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseInteractiveStructure`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
      - :class:`objetto.bases.BaseInteractiveDictStructure`
      - :class:`objetto.bases.BaseInteractiveListStructure`
      - :class:`objetto.bases.BaseInteractiveSetStructure`
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseMutableAuxiliaryStructure(
    BaseAuxiliaryStructure[T],
    BaseMutableStructure[T],
):
    """
    Base mutable auxiliary structure.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseMutableStructure`

    Inherited By:
      - :class:`objetto.bases.BaseMutableAuxiliaryObject`
      - :class:`objetto.bases.BaseMutableDictStructure`
      - :class:`objetto.bases.BaseMutableListStructure`
      - :class:`objetto.bases.BaseMutableSetStructure`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""Dictionary objects and proxy."""

from collections import Counter as ValueCounter
from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, iterkeys, itervalues, raise_from, with_metaclass

from .._applications import Application
from .._bases import FINAL_METHOD_TAG, MISSING, BaseMutableDict, final, init_context
from .._changes import DictUpdate
from .._data import BaseData, DictData, InteractiveDictData
from .._states import DictState
from .._structures import (
    BaseDictStructure,
    BaseDictStructureMeta,
    BaseMutableDictStructure,
    SerializationError,
)
from .bases import (
    DELETED,
    BaseAuxiliaryObject,
    BaseAuxiliaryObjectFunctions,
    BaseAuxiliaryObjectMeta,
    BaseMutableAuxiliaryObject,
    BaseObject,
    BaseProxyObject,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Counter,
        Dict,
        Iterable,
        Iterator,
        Mapping,
        Optional,
        Tuple,
        Type,
        Union,
    )

    from .._applications import Store
    from .._history import HistoryObject

__all__ = ["DictObject", "MutableDictObject", "ProxyDictObject"]


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


@final
class DictObjectFunctions(BaseAuxiliaryObjectFunctions):
    """Static functions for `DictObject`."""

    __slots__ = ()

    @staticmethod
    def make_data_cls_dct(auxiliary_cls):
        # type: (Type[BaseAuxiliaryObject]) -> Dict[str, Any]
        """
        Make data class member dictionary.

        :param auxiliary_cls: Base auxiliary object class.
        :return: Data class member dictionary.
        """
        dct = super(DictObjectFunctions, DictObjectFunctions).make_data_cls_dct(
            auxiliary_cls
        )
        dct.update(
            {
                "_key_relationship": cast(
                    "Type[DictObject]", auxiliary_cls
                )._key_relationship
            }
        )
        return dct

    @staticmethod
    def replace_child_data(store, child, data_location, new_child_data):
        # type: (Store, BaseObject, Any, BaseData) -> Store
        """
        Replace child data.

        :param store: Object's store.
        :param child: Child getting their data replaced.
        :param data_location: Location of the existing child's data.
        :param new_child_data: New child's data.
        :return: Updated object's store.
        """
        original_data = store.data
        assert original_data is not None
        data = cast("DictData", original_data)._set(data_location, new_child_data)
        return store.set("data", data)

    @staticmethod
    def update(
        obj,  # type: DictObject
        input_values,  # type: Mapping
        factory=True,  # type: bool
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        """
        Update values.

        :param obj: Object.
        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :param history: History than triggered this change during undo/redo.
        """
        cls = type(obj)
        relationship = cls._relationship
        key_relationship = cls._key_relationship
        with obj.app.__.write_context(obj) as (read, write):
            if not input_values:
                return

            # Get state, data, and locations cache.
            store = read()
            state = old_state = store.state  # type: DictState
            data = store.data  # type: DictData
            metadata = store.metadata  # type: InteractiveDictData
            locations = metadata.get("locations", DictState())  # type: DictState

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            old_children = set()
            new_children = set()
            history_adopters = set()
            new_values = {}
            old_values = {}

            # Fabricate keys first.
            if factory and cls._key_relationship.factory is not None:
                input_values = dict(
                    (key_relationship.fabricate_key(k, factory=True), v)
                    for k, v in iteritems(input_values)
                )

            # For every input value.
            for key, value in iteritems(input_values):

                # Are we deleting it?
                delete_item = value is DELETED

                # Fabricate new value if not deleting.
                if not delete_item:
                    if factory:
                        value = relationship.fabricate_value(
                            value, factory=factory, **{"app": obj.app}
                        )
                new_values[key] = value

                # Get old value.
                try:
                    old_value = store.state[key]
                except KeyError:
                    if delete_item:
                        error = "can't delete non-existing key '{}'".format(key)
                        raise KeyError(error)
                    old_value = DELETED
                else:
                    if value is old_value:
                        continue
                old_values[key] = old_value

                # Child relationship.
                if relationship.child:
                    same_app = not delete_item and obj._in_same_application(value)

                    # Update children counter, old/new children sets, and locations.
                    if old_value is not DELETED:
                        if obj._in_same_application(old_value):
                            child_counter[old_value] -= 1
                            old_children.add(old_value)
                            locations = locations.remove(old_value)
                    if same_app:
                        child_counter[value] += 1
                        new_children.add(value)
                        locations = locations.set(value, key)

                    # Add history adopter.
                    if relationship.history and same_app:
                        history_adopters.add(value)

                    # Update data.
                    if relationship.data:
                        if delete_item:
                            data = data._remove(key)
                        else:
                            data_relationship = relationship.data_relationship
                            assert data_relationship is not None
                            if same_app:
                                with value.app.__.write_context(value) as (v_read, _):
                                    data = data._set(
                                        key,
                                        data_relationship.fabricate_value(
                                            v_read().data
                                        ),
                                    )
                            else:
                                data = data._set(
                                    key,
                                    data_relationship.fabricate_value(value),
                                )

                # Update state.
                if not delete_item:
                    state = state.set(key, value)
                else:
                    state = state.remove(key)

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = DictUpdate(
                __redo__=DictObjectFunctions.redo_update,
                __undo__=DictObjectFunctions.undo_update,
                obj=obj,
                old_children=old_children,
                new_children=new_children,
                old_values=old_values,
                new_values=new_values,
                old_state=old_state,
                new_state=state,
                history_adopters=history_adopters,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_update(change):
        # type: (DictUpdate) -> None
        """
        Redo update values.

        :param change: Dictionary update changes.
        """
        DictObjectFunctions.update(
            cast("DictObject", change.obj),
            change.new_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def undo_update(change):
        # type: (DictUpdate) -> None
        """
        Undo update values.

        :param change: Dictionary update changes.
        """
        DictObjectFunctions.update(
            cast("DictObject", change.obj),
            change.old_values,
            factory=False,
            history=change.obj._history,
        )


type.__setattr__(cast(type, DictObjectFunctions), FINAL_METHOD_TAG, True)


class DictObjectMeta(BaseAuxiliaryObjectMeta, BaseDictStructureMeta):
    """Metaclass for `DictObject`."""

    @property
    @final
    def _state_factory(cls):
        # type: () -> Callable[..., DictState]
        """State factory."""
        return DictState

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[DictObject]
        """Base auxiliary object type."""
        return DictObject

    @property
    @final
    def _base_auxiliary_data_type(cls):
        # type: () -> Type[DictData]
        """Base auxiliary data type."""
        return DictData


# noinspection PyTypeChecker
_DO = TypeVar("_DO", bound="DictObject")


class DictObject(
    with_metaclass(
        DictObjectMeta,
        BaseAuxiliaryObject[KT],
        BaseDictStructure[KT, VT],
    )
):
    """
    Dictionary object.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryObject`
      - :class:`objetto.bases.BaseDictStructure`

    Inherited by:
      - :class:`objetto.objects.MutableDictObject`

    :param app: Application.
    :type app: objetto.applications.Application

    :param initial: Initial values.
    :type initial: collections.abc.Mapping or collections.abc.Iterable[\
tuple[collections.abc.Hashable, Any]]
    """

    __slots__ = ()
    __functions__ = DictObjectFunctions

    def __init__(
        self,
        app,  # type: Application
        initial=(),  # type: Union[Mapping[KT, VT], Iterable[Tuple[KT, VT]]]
    ):
        # type: (...) -> None
        super(DictObject, self).__init__(app=app)
        self.__functions__.update(self, dict(initial))

    @final
    def _clear(self):
        # type: (_DO) -> _DO
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.objects.DictObject

        :raises AttributeError: No deletable attributes.
        """
        with self.app.write_context():
            self._update((k, DELETED) for k in self._state)
        return self

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DO, Mapping[KT, VT], VT) -> _DO
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DO, Iterable[Tuple[KT, VT]], VT) -> _DO
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_DO, VT) -> _DO
        pass

    @final
    def _update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.objects.DictObject
        """
        self.__functions__.update(self, dict(*args, **kwargs))
        return self

    @final
    def _set(self, key, value):
        # type: (_DO, KT, VT) -> _DO
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.DictObject
        """
        self.__functions__.update(self, {key: value})
        return self

    @final
    def _discard(self, key):
        # type: (_DO, KT) -> _DO
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.DictObject
        """
        with self.app.write_context():
            if key in self._state:
                self.__functions__.update(self, {key: DELETED})
        return self

    @final
    def _remove(self, key):
        # type: (_DO, KT) -> _DO
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.DictObject

        :raises KeyError: Key is not present.
        """
        self.__functions__.update(self, {key: DELETED})
        return self

    @final
    def _locate(self, child):
        # type: (BaseObject) -> KT
        """
        Locate child object.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Location.
        :rtype: collections.abc.Hashable

        :raises ValueError: Could not locate child.
        """
        with self.app.__.read_context(self) as read:
            metadata = read().metadata
            try:
                return metadata["locations"][child]
            except KeyError:
                error = "could not locate child {} in {}".format(child, self)
                exc = ValueError(error)
                raise_from(exc, None)
                raise exc

    @final
    def _locate_data(self, child):
        # type: (BaseObject) -> KT
        """
        Locate child object's data.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Data location.
        :rtype: collections.abc.Hashable

        :raises ValueError: Could not locate child's data.
        """
        return self._locate(child)

    @classmethod
    @final
    def deserialize(cls, serialized, app=None, **kwargs):
        # type: (Type[_DO], Dict[str, Any], Application, Any) -> _DO
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: dict

        :param app: Application (required).
        :type app: objetto.applications.Application

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.objects.DictObject

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        if app is None:
            error = (
                "missing required 'app' keyword argument for '{}.deserialize()' method"
            ).format(cls.__fullname__)
            raise ValueError(error)
        kwargs["app"] = app

        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)

        with app.write_context():
            self = cast("_DO", cls.__new__(cls))
            with init_context(self):
                super(DictObject, self).__init__(app)
                initial = dict(
                    (n, cls.deserialize_value(v, None, **kwargs))
                    for n, v in iteritems(serialized)
                    if cls._relationship.serialized
                )
                self.__functions__.update(self, initial)
            return self

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> Dict
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to serializer functions.

        :return: Serialized.
        :rtype: dict

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        with self.app.read_context():
            if not type(self)._relationship.serialized:
                error = "'{}' is not serializable".format(type(self).__fullname__)
                raise SerializationError(error)

            return dict(
                (k, self.serialize_value(v, None, **kwargs))
                for (k, v) in iteritems(self._state)
                if type(self)._relationship.serialized
            )

    @property
    @final
    def _state(self):
        # type: () -> DictState[KT, VT]
        """
        State.

        :rtype: objetto.states.DictState
        """
        return cast("DictState[KT, VT]", super(BaseDictStructure, self)._state)

    @property
    @final
    def data(self):
        # type: () -> DictData[KT, VT]
        """
        Data.

        :rtype: objetto.data.DictData
        """
        return cast("DictData[KT, VT]", super(BaseDictStructure, self).data)


# noinspection PyAbstractClass
class MutableDictObject(
    BaseMutableDictStructure[KT, VT], DictObject[KT, VT], BaseMutableAuxiliaryObject[KT]
):
    """
    Mutable dictionary object.

    Inherits from:
      - :class:`objetto.bases.BaseMutableDictStructure`
      - :class:`objetto.objects.DictObject`
      - :class:`objetto.bases.BaseMutableAuxiliaryObject`
    """

    __slots__ = ()

    @final
    def pop(self, key, fallback=MISSING):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key and remove it, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.

        :raises KeyError: Key is not present and fallback value not provided.
        """
        with self.app.write_context():
            try:
                value = self[key]
            except KeyError:
                if fallback is not MISSING:
                    return fallback
                raise
            else:
                self._remove(key)
                return value

    @final
    def popitem(self):
        # type: () -> Tuple[KT, VT]
        """
        Get item and discard key.

        :return: Item.
        :rype: tuple[collections.abc.Hashable, Any]

        :raises KeyError: Dictionary is empty.
        """
        with self.app.write_context():
            if not self:
                error = "dictionary is empty"
                raise KeyError(error)
            key = next(iter(self))
            return key, self.pop(key)

    @final
    def setdefault(self, key, default=None):
        # type: (KT, Optional[VT]) -> VT
        """
        Get the value for the specified key, insert key with default if not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param default: Default value.

        :return: Existing or default value.
        """
        with self.app.write_context():
            try:
                return self[key]
            except KeyError:
                self._set(key, cast("VT", default))
                return self[key]


# noinspection PyTypeChecker
_PDO = TypeVar("_PDO", bound="ProxyDictObject")


@final
class ProxyDictObject(BaseProxyObject[KT], BaseMutableDict[KT, VT]):
    """
    Mutable proxy dictionary.

    Inherits from:
      - :class:`objetto.bases.BaseProxyObject`
      - :class:`objetto.bases.BaseMutableDict`
    """

    __slots__ = ()

    def __reversed__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over reversed keys.

        :return: Reversed keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        return reversed(self._state)

    def __getitem__(self, key):
        # type: (KT) -> VT
        """
        Get value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Value.

        :raises KeyError: Invalid key.
        """
        return self._obj.__getitem__(key)

    @overload
    def _update(self, __m, **kwargs):
        # type: (_PDO, Mapping[KT, VT], VT) -> _PDO
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_PDO, Iterable[Tuple[KT, VT]], VT) -> _PDO
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_PDO, VT) -> _PDO
        pass

    def _update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.objects.ProxyDictObject
        """
        self._obj._update(*args, **kwargs)
        return self

    def _set(self, key, value):
        # type: (_PDO, KT, VT) -> _PDO
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.ProxyDictObject
        """
        self._obj._set(key, value)
        return self

    def _discard(self, key):
        # type: (_PDO, KT) -> _PDO
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxyDictObject
        """
        self._obj._remove(key)
        return self

    def _remove(self, key):
        # type: (_PDO, KT) -> _PDO
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxyDictObject

        :raises KeyError: Key is not present.
        """
        self._obj._remove(key)
        return self

    def get(self, key, fallback=None):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.
        """
        return self._obj.get(key, fallback)

    def iteritems(self):
        # type: () -> Iterator[Tuple[KT, VT]]
        """
        Iterate over items.

        :return: Item iterator.
        :rtype: collections.abc.Iterator[tuple[collections.abc.Hashable, Any]]
        """
        for key, value in iteritems(self._state):
            yield key, value

    def iterkeys(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in iterkeys(self._state):
            yield key

    def itervalues(self):
        # type: () -> Iterator[VT]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator
        """
        for value in itervalues(self._state):
            yield value

    def pop(self, key, fallback=MISSING):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key and remove it, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.

        :raises KeyError: Key is not present and fallback value not provided.
        """
        with self.app.write_context():
            try:
                value = self[key]
            except KeyError:
                if fallback is not MISSING:
                    return fallback
                raise
            else:
                self._remove(key)
                return value

    def popitem(self):
        # type: () -> Tuple[KT, VT]
        """
        Get item and discard key.

        :return: Item.
        :rtype: tuple[collections.abc.Hashable, Any]

        :raises KeyError: Dictionary is empty.
        """
        with self.app.write_context():
            if not self:
                error = "dictionary is empty"
                raise KeyError(error)
            key = next(iter(self))
            return key, self.pop(key)

    def setdefault(self, key, default=None):
        # type: (KT, Optional[VT]) -> VT
        """
        Get the value for the specified key, insert key with default if not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param default: Default value.

        :return: Existing or default value.
        """
        with self.app.write_context():
            try:
                return self[key]
            except KeyError:
                self._set(key, cast("VT", default))
                return self[key]

    @property
    def _obj(self):
        # type: () -> DictObject[KT, VT]
        """
        Dict obje

        :rtype: objetto.objects.DictObject
        """
        return cast("DictObject[KT, VT]", super(ProxyDictObject, self)._obj)

    @property
    def _state(self):
        # type: () -> DictState[KT, VT]
        """
        State.

        :rtype: objetto.states.DictState
        """
        return cast("DictState[KT, VT]", super(ProxyDictObject, self)._state)

    @property
    def data(self):
        # type: () -> Optional[DictData[KT, VT]]
        """
        Data.

        :rtype: objetto.data.DictData or None
        """
        return cast("Optional[DictData[KT, VT]]", super(ProxyDictObject, self).data)



# -*- coding: utf-8 -*-
"""List objects and proxy."""

from collections import Counter as ValueCounter
from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import raise_from, with_metaclass

from .._applications import Application
from .._bases import FINAL_METHOD_TAG, BaseMutableList, final, init_context
from .._changes import ListDelete, ListInsert, ListMove, ListUpdate
from .._data import BaseData, InteractiveDictData, ListData
from .._states import ListState
from .._structures import (
    BaseListStructure,
    BaseListStructureMeta,
    BaseMutableListStructure,
    SerializationError,
)
from ..utils.list_operations import pre_move, resolve_continuous_slice, resolve_index
from .bases import (
    BaseAuxiliaryObject,
    BaseAuxiliaryObjectFunctions,
    BaseAuxiliaryObjectMeta,
    BaseMutableAuxiliaryObject,
    BaseObject,
    BaseProxyObject,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Counter,
        Dict,
        Iterable,
        Iterator,
        List,
        MutableSequence,
        Optional,
        Sequence,
        Set,
        Tuple,
        Type,
        Union,
    )

    from .._applications import Store
    from .._history import HistoryObject

__all__ = ["ListObject", "MutableListObject", "ProxyListObject"]


T = TypeVar("T")  # Any type.


@final
class ListObjectFunctions(BaseAuxiliaryObjectFunctions):
    """Static functions for `ListObject`."""

    __slots__ = ()

    @staticmethod
    def replace_child_data(store, child, data_location, new_child_data):
        # type: (Store, BaseObject, Any, BaseData) -> Store
        """
        Replace child data.

        :param store: Object's store.
        :param child: Child getting their data replaced.
        :param data_location: Location of the existing child's data.
        :param new_child_data: New child's data.
        :return: Updated object's store.
        """
        original_data = store.data
        assert original_data is not None
        data = cast("ListData", original_data)._update(data_location, new_child_data)
        return store.set("data", data)

    @staticmethod
    def insert(
        obj,  # type: ListObject
        index,  # type: int
        input_values,  # type: Iterable[Any]
        factory=True,  # type: bool
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        cls = type(obj)
        relationship = cls._relationship
        with obj.app.__.write_context(obj) as (read, write):
            if not input_values:
                return

            # Get state, data, and a brand new locations cache.
            store = read()
            state = old_state = store.state  # type: ListState
            data = store.data  # type: ListData
            metadata = store.metadata  # type: InteractiveDictData
            locations = {}  # type: Dict[BaseObject, int]

            # Get resolved index.
            index = resolve_index(len(state), index, clamp=True)

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            new_children = set()  # type: Set[BaseObject]
            history_adopters = set()  # type: Set[BaseObject]
            new_values = []  # type: List[Any]
            new_data_values = []  # type: List[Any]

            # For every input value.
            for i, value in enumerate(input_values):

                # Fabricate new value.
                if factory:
                    value = relationship.fabricate_value(
                        value, factory=factory, **{"app": obj.app}
                    )
                new_values.append(value)

                # Child relationship.
                if relationship.child:
                    same_app = obj._in_same_application(value)

                    # Update children counter and new children set.
                    if same_app:
                        child_counter[value] += 1
                        new_children.add(value)
                        locations[value] = index + i

                    # Add history adopter.
                    if relationship.history and same_app:
                        history_adopters.add(value)

                    # Update data.
                    if relationship.data:
                        data_relationship = relationship.data_relationship
                        assert data_relationship is not None
                        if same_app:
                            with value.app.__.write_context(value) as (v_read, _):
                                data_value = data_relationship.fabricate_value(
                                    v_read().data,
                                )
                        else:
                            data_value = data_relationship.fabricate_value(value)
                        new_data_values.append(data_value)

            # Update state and data.
            state = state.insert(index, *new_values)
            if relationship.data:
                data = data._insert(index, *new_data_values)

            # Get last index and stop.
            stop = index + len(new_values)
            last_index = stop - 1

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = ListInsert(
                __redo__=ListObjectFunctions.redo_insert,
                __undo__=ListObjectFunctions.undo_insert,
                obj=obj,
                old_children=(),
                new_children=new_children,
                index=index,
                last_index=last_index,
                stop=stop,
                new_values=new_values,
                old_state=old_state,
                new_state=state,
                history_adopters=history_adopters,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_insert(change):
        # type: (ListInsert) -> None
        ListObjectFunctions.insert(
            cast("ListObject", change.obj),
            change.index,
            change.new_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def undo_insert(change):
        # type: (ListInsert) -> None
        ListObjectFunctions.delete(
            cast("ListObject", change.obj),
            slice(change.index, change.stop),
            history=change.obj._history,
        )

    @staticmethod
    def delete(
        obj,  # type: ListObject
        item,  # type: Union[int, slice]
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        cls = type(obj)
        relationship = cls._relationship
        with obj.app.__.write_context(obj) as (read, write):

            # Get state, data, and a brand new locations cache.
            store = read()
            state = old_state = store.state  # type: ListState
            data = store.data  # type: ListData
            metadata = store.metadata  # type: InteractiveDictData
            locations = {}  # type: Dict[BaseObject, int]

            # Get resolved indexes and stop.
            if isinstance(item, slice):
                index, stop = resolve_continuous_slice(len(state), item)
                if stop == index:
                    return
            else:
                index = resolve_index(len(state), item)
                stop = index + 1
            last_index = stop - 1
            slc = slice(index, stop)

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            old_children = set()  # type: Set[BaseObject]
            old_values = state[index : last_index + 1]  # type: ListState

            # For every value being removed.
            for value in old_values:

                # Child relationship.
                if relationship.child:
                    same_app = obj._in_same_application(value)

                    # Update children counter and new children set.
                    if same_app:
                        child_counter[value] -= 1
                        old_children.add(value)

            # Update state and data.
            state = state.delete(slc)
            if relationship.data:
                data = data._delete(slc)

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = ListDelete(
                __redo__=ListObjectFunctions.redo_delete,
                __undo__=ListObjectFunctions.undo_delete,
                obj=obj,
                old_children=old_children,
                new_children=(),
                index=index,
                last_index=last_index,
                stop=stop,
                old_values=old_values,
                old_state=old_state,
                new_state=state,
                history_adopters=(),
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_delete(change):
        # type: (ListDelete) -> None
        ListObjectFunctions.delete(
            cast("ListObject", change.obj),
            slice(change.index, change.stop),
            history=change.obj._history,
        )

    @staticmethod
    def undo_delete(change):
        # type: (ListDelete) -> None
        ListObjectFunctions.insert(
            cast("ListObject", change.obj),
            change.index,
            change.old_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def update(
        obj,  # type: ListObject
        item,  # type: Union[int, slice]
        input_values,  # type: Iterable[Any]
        factory=True,  # type: bool
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        cls = type(obj)
        relationship = cls._relationship
        with obj.app.__.write_context(obj) as (read, write):

            # Get state, data, and a brand new locations cache.
            store = read()
            state = old_state = store.state  # type: ListState
            data = store.data  # type: ListData
            metadata = store.metadata  # type: InteractiveDictData
            locations = {}  # type: Dict[BaseObject, int]

            # Get old values and check length.
            if isinstance(item, slice):
                index, stop = resolve_continuous_slice(len(state), item)
            else:
                index = resolve_index(len(state), item)
                stop = index + 1
            slc = slice(index, stop)
            last_index = stop - 1
            input_values = list(input_values)
            old_values = state[slc]
            if len(old_values) != len(input_values):
                error = "length of slice and values mismatch"
                raise IndexError(error)
            if len(old_values) == 0:
                return

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            history_adopters = set()
            old_children = set()
            new_children = set()
            new_values = []
            new_data_values = []

            # For every value being removed.
            for value, old_value in zip(input_values, old_values):

                # Fabricate new value.
                if factory:
                    value = relationship.fabricate_value(
                        value, factory=factory, **{"app": obj.app}
                    )
                new_values.append(value)

                # No change.
                if value is old_value:
                    continue

                # Child relationship.
                if relationship.child:
                    same_app = obj._in_same_application(value)

                    # Update children counter, old/new children sets.
                    if obj._in_same_application(old_value):
                        child_counter[old_value] -= 1
                        old_children.add(old_value)
                    if same_app:
                        child_counter[value] += 1
                        new_children.add(value)

                    # Add history adopter.
                    if relationship.history and same_app:
                        history_adopters.add(value)

                    # Update data.
                    if relationship.data:
                        data_relationship = relationship.data_relationship
                        assert data_relationship is not None
                        if same_app:
                            with value.app.__.write_context(value) as (v_read, _):
                                data_value = data_relationship.fabricate_value(
                                    v_read().data,
                                )
                        else:
                            data_value = data_relationship.fabricate_value(value)
                        new_data_values.append(data_value)

            # Update state and data.
            state = state.update(index, *new_values)
            if relationship.data:
                data = data._update(index, *new_data_values)

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = ListUpdate(
                __redo__=ListObjectFunctions.redo_update,
                __undo__=ListObjectFunctions.undo_update,
                obj=obj,
                old_children=old_children,
                new_children=new_children,
                history_adopters=history_adopters,
                index=index,
                last_index=last_index,
                stop=stop,
                old_values=old_values,
                new_values=new_values,
                old_state=old_state,
                new_state=state,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_update(change):
        # type: (ListUpdate) -> None
        ListObjectFunctions.update(
            cast("ListObject", change.obj),
            slice(change.index, change.stop),
            change.new_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def undo_update(change):
        # type: (ListUpdate) -> None
        ListObjectFunctions.update(
            cast("ListObject", change.obj),
            slice(change.index, change.stop),
            change.old_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def move(
        obj,  # type: ListObject
        item,  # type: Union[int, slice]
        target_index,  # type: int
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        with obj.app.__.write_context(obj) as (read, write):

            # Get state, data, and a brand new locations cache.
            store = read()
            state = old_state = store.state  # type: ListState
            data = store.data  # type: ListData
            metadata = store.metadata  # type: InteractiveDictData
            locations = {}  # type: Dict[BaseObject, int]

            # Get resolved indexes and stop.
            pre_move_result = pre_move(len(state), item, target_index)
            if pre_move_result is None:
                return
            index, stop, target_index, post_index = pre_move_result
            post_stop = post_index + (stop - index)
            last_index = stop - 1
            post_last_index = post_stop - 1

            # Prepare change information.
            values = state[index : last_index + 1]

            # Update state and data.
            state = state.move(item, target_index)
            data = data._move(item, target_index)

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = ListMove(
                __redo__=ListObjectFunctions.redo_move,
                __undo__=ListObjectFunctions.undo_move,
                obj=obj,
                old_children=(),
                new_children=(),
                history_adopters=(),
                index=index,
                last_index=last_index,
                stop=stop,
                target_index=target_index,
                post_index=post_index,
                post_last_index=post_last_index,
                post_stop=post_stop,
                values=values,
                old_state=old_state,
                new_state=state,
                history=history,
            )
            write(state, data, metadata, ValueCounter(), change)

    @staticmethod
    def redo_move(change):
        # type: (ListMove) -> None
        ListObjectFunctions.move(
            cast("ListObject", change.obj),
            slice(change.index, change.stop),
            change.target_index,
            history=change.obj._history,
        )

    @staticmethod
    def undo_move(change):
        # type: (ListMove) -> None
        ListObjectFunctions.move(
            cast("ListObject", change.obj),
            slice(change.post_index, change.post_stop),
            change.index,
            history=change.obj._history,
        )


type.__setattr__(cast(type, ListObjectFunctions), FINAL_METHOD_TAG, True)


class ListObjectMeta(BaseAuxiliaryObjectMeta, BaseListStructureMeta):
    """Metaclass for `ListObject`."""

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(ListObjectMeta, cls).__init__(name, bases, dct)

        # Prevent subclasses from overriding '__getitem__'.
        if "__getitem__" in dct:
            allowed_classes = set()
            try:
                allowed_classes.add(ListObject)
                allowed_classes.add(MutableListObject)
            except NameError:
                pass
            if len(allowed_classes) == 2 and cls not in allowed_classes:
                error = "can't override final member '__getitem__'"
                raise TypeError(error)

    @property
    @final
    def _state_factory(cls):
        # type: () -> Callable[..., ListState]
        """State factory."""
        return ListState

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[ListObject]
        """Base auxiliary object type."""
        return ListObject

    @property
    @final
    def _base_auxiliary_data_type(cls):
        # type: () -> Type[ListData]
        """Base auxiliary data type."""
        return ListData


# noinspection PyTypeChecker
_LO = TypeVar("_LO", bound="ListObject")


class ListObject(
    with_metaclass(
        ListObjectMeta,
        BaseAuxiliaryObject[T],
        BaseListStructure[T],
    )
):
    """
    List object.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryObject`
      - :class:`objetto.bases.BaseListStructure`

    Inherited by:
      - :class:`objetto.objects.MutableListObject`

    :param app: Application.
    :type app: objetto.applications.Application

    :param initial: Initial values.
    :type initial: collections.abc.Iterable
    """

    __slots__ = ()
    __functions__ = ListObjectFunctions

    def __init__(self, app, initial=()):
        # type: (Application, Iterable[T]) -> None
        super(ListObject, self).__init__(app=app)
        self.__functions__.insert(self, 0, initial)

    @overload
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    def __getitem__(self, index):
        # type: (slice) -> Sequence[T]
        pass

    # @final (special case, taken care of by the metaclass)
    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: Any or objetto.states.ListState
        """
        return self._state[index]

    @final
    def _clear(self):
        # type: (_LO) -> _LO
        """
        Clear all values.

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        with self.app.write_context():
            state_length = len(self._state)
            if state_length:
                self._delete(slice(0, state_length))
        return self

    @final
    def _insert(self, index, *values):
        # type: (_LO, int, T) -> _LO
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.objects.ListObject

        :raises ValueError: No values provided.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        self.__functions__.insert(self, index, values)
        return self

    @final
    def _append(self, value):
        # type: (_LO, T) -> _LO
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        self.__functions__.insert(self, len(self._state), (value,))
        return self

    @final
    def _extend(self, iterable):
        # type: (_LO, Iterable[T]) -> _LO
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        self.__functions__.insert(self, len(self._state), iterable)
        return self

    @final
    def _remove(self, value):
        # type: (_LO, T) -> _LO
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.ListObject

        :raises ValueError: Value is not present.
        """
        with self.app.write_context():
            index = self.index(value)
            self.__functions__.delete(self, index)
            return self

    @final
    def _reverse(self):
        # type: (_LO) -> _LO
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        with self.app.write_context():
            if self._state:
                reversed_values = self._state.reverse()
                self.__functions__.update(
                    self, slice(0, len(self._state)), reversed_values
                )
        return self

    @final
    def _move(self, item, target_index):
        # type: (_LO, Union[slice, int], int) -> _LO
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        self.__functions__.move(self, item, target_index)
        return self

    @final
    def _delete(self, item):
        # type: (_LO, Union[slice, int]) -> _LO
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.objects.ListObject
        """
        self.__functions__.delete(self, item)
        return self

    @final
    def _update(self, index, *values):
        # type: (_LO, int, T) -> _LO
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.objects.ListObject

        :raises ValueError: No values provided.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        self.__functions__.update(self, index, values)
        return self

    @final
    def _locate(self, child):
        # type: (BaseObject) -> int
        """
        Locate child object.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Location.
        :rtype: int

        :raises ValueError: Could not locate child.
        """
        with self.app.__.read_context(self) as read:
            metadata = read().metadata
            try:
                return metadata["locations"][child]
            except KeyError:
                if child in self._children:
                    location = metadata["locations"][child] = self.index(child)
                    return location
                error = "could not locate child {} in {}".format(child, self)
                exc = ValueError(error)
                raise_from(exc, None)
                raise exc

    @final
    def _locate_data(self, child):
        # type: (BaseObject) -> int
        """
        Locate child object's data.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Data location.
        :rtype: int

        :raises ValueError: Could not locate child's data.
        """
        return self._locate(child)

    @classmethod
    @final
    def deserialize(cls, serialized, app=None, **kwargs):
        # type: (Type[_LO], Dict[str, Any], Application, Any) -> _LO
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: list

        :param app: Application (required).
        :type app: objetto.applications.Application

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.objects.ListObject

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        if app is None:
            error = (
                "missing required 'app' keyword argument for '{}.deserialize()' method"
            ).format(cls.__fullname__)
            raise ValueError(error)
        kwargs["app"] = app

        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)

        with app.write_context():
            self = cast("_LO", cls.__new__(cls))
            with init_context(self):
                super(ListObject, self).__init__(app)
                initial = (
                    cls.deserialize_value(v, None, **kwargs)
                    for v in serialized
                    if cls._relationship.serialized
                )
                self.__functions__.insert(self, 0, initial)
            return self

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> List
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: list

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        with self.app.read_context():

            if not type(self)._relationship.serialized:
                error = "'{}' is not serializable".format(type(self).__fullname__)
                raise SerializationError(error)

            return list(
                self.serialize_value(v, None, **kwargs)
                for v in self._state
                if type(self)._relationship.serialized
            )

    @property
    @final
    def _state(self):
        # type: () -> ListState[T]
        """
        State.

        :rtype: objetto.states.ListState
        """
        return cast("ListState[T]", super(BaseListStructure, self)._state)

    @property
    @final
    def data(self):
        # type: () -> ListData[T]
        """
        Data.

        :rtype: objetto.data.ListData
        """
        return cast("ListData[T]", super(BaseListStructure, self).data)


# noinspection PyAbstractClass
class MutableListObject(
    BaseMutableListStructure[T], ListObject[T], BaseMutableAuxiliaryObject[T]
):
    """
    Mutable list object.

    Inherits from:
      - :class:`objetto.bases.BaseMutableListStructure`
      - :class:`objetto.objects.ListObject`
      - :class:`objetto.bases.BaseMutableAuxiliaryObject`

    .. code:: python

        >>> from objetto import Application, Object, attribute
        >>> from objetto.objects import MutableListObject, Relationship

        >>> class Hobby(Object):
        ...     description = attribute(str)
        ...
        >>> class HobbiesList(MutableListObject):  # inherit from MutableListObject
        ...     _relationship = Relationship(Hobby)  # define relationship with type
        ...
        >>> app = Application()
        >>> hobby_a = Hobby(app, description="biking")
        >>> hobby_b = Hobby(app, description="gaming")
        >>> hobbies = HobbiesList(app)  # make new instance
        >>> hobbies.extend((hobby_a, hobby_b))  # extend list with 'hobby' objects
    """

    __slots__ = ()

    @overload
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    def __getitem__(self, index):
        # type: (slice) -> MutableSequence[T]
        pass

    @final
    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: Any or list
        """
        if isinstance(index, slice):
            return list(self._state[index])
        else:
            return self._state[index]

    @overload
    def __setitem__(self, index, value):
        # type: (int, T) -> None
        pass

    @overload
    def __setitem__(self, slc, values):
        # type: (slice, Iterable[T]) -> None
        pass

    @final
    def __setitem__(self, item, value):
        # type: (Union[int, slice], Union[T, Iterable[T]]) -> None
        """
        Set value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :param value: Value/values.
        :type value: Any or collections.abc.Iterable

        :raises IndexError: Slice is noncontinuous.
        :raises ValueError: Values length does not fit in slice.
        """
        if isinstance(item, slice):
            with self.app.write_context():
                values = tuple(cast("Iterable[T]", value))
                index, stop = self.resolve_continuous_slice(item)
                if len(values) != stop - index:
                    error = "values length ({}) does not fit in slice ({})".format(
                        len(values), stop - index
                    )
                    raise ValueError(error)
                self._update(index, *values)
        else:
            self._update(item, cast("T", value))

    @overload
    def __delitem__(self, index):
        # type: (int) -> None
        pass

    @overload
    def __delitem__(self, slc):
        # type: (slice) -> None
        pass

    @final
    def __delitem__(self, item):
        # type: (Union[int, slice]) -> None
        """
        Delete value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :raises IndexError: Slice is noncontinuous.
        """
        self._delete(item)

    @final
    def pop(self, index=-1):
        # type: (int) -> T
        """
        Pop value from index.

        :param index: Index.
        :type index: int

        :return: Value.
        """
        with self.app.write_context():
            value = self[index]
            self._delete(index)
        return value


# noinspection PyTypeChecker
_PLO = TypeVar("_PLO", bound="ProxyListObject")


@final
class ProxyListObject(BaseProxyObject[T], BaseMutableList[T]):
    """
    Mutable proxy list.

    Inherits from:
      - :class:`objetto.bases.BaseProxyObject`
      - :class:`objetto.bases.BaseMutableList`
    """

    __slots__ = ()

    @overload
    def __setitem__(self, index, value):
        # type: (int, T) -> None
        pass

    @overload
    def __setitem__(self, slc, values):
        # type: (slice, Iterable[T]) -> None
        pass

    def __setitem__(self, item, value):
        # type: (Union[int, slice], Union[T, Iterable[T]]) -> None
        """
        Set value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :param value: Value/values.
        :type value: Any or collections.abc.Iterable

        :raises IndexError: Slice is noncontinuous.
        :raises ValueError: Values length does not fit in slice.
        """
        if isinstance(item, slice):
            with self.app.write_context():
                values = tuple(cast("Iterable[T]", value))
                index, stop = self.resolve_continuous_slice(item)
                if len(values) != stop - index:
                    error = "values length ({}) does not fit in slice ({})".format(
                        len(values), stop - index
                    )
                    raise ValueError(error)
                self._update(index, *values)
        else:
            self._update(item, cast("T", value))

    @overload
    def __delitem__(self, index):
        # type: (int) -> None
        pass

    @overload
    def __delitem__(self, slc):
        # type: (slice) -> None
        pass

    def __delitem__(self, item):
        # type: (Union[int, slice]) -> None
        """
        Delete value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :raises IndexError: Slice is noncontinuous.
        """
        self._delete(item)

    def __reversed__(self):
        # type: () -> Iterator[T]
        """
        Iterate over reversed values.

        :return: Reversed values iterator.
        :rtype: collections.abc.Iterator
        """
        return reversed(self._state)

    @overload
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    def __getitem__(self, index):
        # type: (slice) -> MutableSequence[T]
        pass

    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: list
        """
        if isinstance(index, slice):
            return list(self._state[index])
        else:
            return self._obj[index]

    def _insert(self, index, *values):
        # type: (_PLO, int, T) -> _PLO
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject

        :raises ValueError: No values provided.
        """
        self._obj._insert(index, *values)
        return self

    def _append(self, value):
        # type: (_PLO, T) -> _PLO
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject
        """
        self._obj._append(value)
        return self

    def _extend(self, iterable):
        # type: (_PLO, Iterable[T]) -> _PLO
        """
        Extend at the end with iterable.

        :param iterable: Iterable.

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject
        """
        self._obj._extend(iterable)
        return self

    def _remove(self, value):
        # type: (_PLO, T) -> _PLO
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject

        :raises ValueError: Value is not present.
        """
        self._obj._remove(value)
        return self

    def _reverse(self):
        # type: (_PLO) -> _PLO
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject
        """
        self._obj._reverse()
        return self

    def _move(self, item, target_index):
        # type: (_PLO, Union[slice, int], int) -> _PLO
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject
        """
        self._obj._move(item, target_index)
        return self

    def _delete(self, item):
        # type: (_PLO, Union[slice, int]) -> _PLO
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.objects.ProxyListObject
        """
        self._obj._delete(item)
        return self

    def _update(self, index, *values):
        # type: (_PLO, int, T) -> _PLO
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :raises ValueError: No values provided.
        """
        self._obj._update(index, *values)
        return self

    def pop(self, index=-1):
        # type: (int) -> T
        """
        Pop value from index.

        :param index: Index.
        :type index: int

        :return: Value.
        """
        with self.app.write_context():
            value = self[index]
            self._delete(index)
        return value

    def count(self, value):
        # type: (Any) -> int
        """
        Count number of occurrences of a value.

        :param value: Value.

        :return: Number of occurrences.
        :rtype: int
        """
        return self._obj.count(value)

    def index(self, value, start=None, stop=None):
        # type: (Any, Optional[int], Optional[int]) -> int
        """
        Get index of a value.

        :param value: Value.

        :param start: Start index.
        :type start: int or None

        :param stop: Stop index.
        :type stop: int or None

        :return: Index of value.
        :rtype: int

        :raises ValueError: Provided stop but did not provide start.
        """
        return self._obj.index(value, start=start, stop=stop)

    def resolve_index(self, index, clamp=False):
        # type: (int, bool) -> int
        """
        Resolve index to a positive number.

        :param index: Input index.
        :type index: int

        :param clamp: Whether to clamp between zero and the length.
        :type clamp: bool

        :return: Resolved index.
        :rtype: int

        :raises IndexError: Index out of range.
        """
        return self._obj.resolve_index(index, clamp=clamp)

    def resolve_continuous_slice(self, slc):
        # type: (slice) -> Tuple[int, int]
        """
        Resolve continuous slice according to length.

        :param slc: Continuous slice.
        :type slc: slice

        :return: Index and stop.
        :rtype: tuple[int, int]

        :raises IndexError: Slice is noncontinuous.
        """
        return self._obj.resolve_continuous_slice(slc)

    @property
    def _obj(self):
        # type: () -> ListObject[T]
        """
        List object.

        :rtype: objetto.objects.ListObject
        """
        return cast("ListObject[T]", super(ProxyListObject, self)._obj)

    @property
    def _state(self):
        # type: () -> ListState[T]
        """
        State.

        :rtype: objetto.states.ListState
        """
        return cast("ListState[T]", super(ProxyListObject, self)._state)

    @property
    def data(self):
        # type: () -> Optional[ListData[T]]
        """
        Data.

        :rtype: objetto.data.ListData or None
        """
        return cast("Optional[ListData[T]]", super(ProxyListObject, self).data)



# -*- coding: utf-8 -*-

from .applications import Application, root
from .constants import POST, PRE
from .data import (
    Data,
    InteractiveData,
    data_attribute,
    data_constant_attribute,
    data_dict_attribute,
    data_dict_cls,
    data_list_attribute,
    data_list_cls,
    data_set_attribute,
    data_set_cls,
)
from .objects import (
    Object,
    attribute,
    constant_attribute,
    data_method,
    data_relationship,
    dict_attribute,
    dict_cls,
    history_descriptor,
    list_attribute,
    list_cls,
    protected_attribute_pair,
    protected_dict_attribute_pair,
    protected_list_attribute_pair,
    protected_set_attribute_pair,
    set_attribute,
    set_cls,
    unique_descriptor,
)
from .observers import ActionObserver
from .reactions import reaction

__all__ = [
    "PRE",
    "POST",
    "Application",
    "root",
    "Object",
    "unique_descriptor",
    "history_descriptor",
    "data_relationship",
    "data_method",
    "attribute",
    "protected_attribute_pair",
    "dict_attribute",
    "protected_dict_attribute_pair",
    "list_attribute",
    "protected_list_attribute_pair",
    "set_attribute",
    "protected_set_attribute_pair",
    "dict_cls",
    "list_cls",
    "set_cls",
    "constant_attribute",
    "reaction",
    "ActionObserver",
    "Data",
    "InteractiveData",
    "data_attribute",
    "data_constant_attribute",
    "data_dict_attribute",
    "data_list_attribute",
    "data_set_attribute",
    "data_dict_cls",
    "data_list_cls",
    "data_set_cls",
]



# -*- coding: utf-8 -*-
"""Base dictionary classes."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from slotted import SlottedMapping, SlottedMutableMapping

from .bases import (
    MISSING,
    BaseCollection,
    BaseInteractiveCollection,
    BaseMutableCollection,
    BaseProtectedCollection,
    Generic,
    final,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        ItemsView,
        Iterable,
        Iterator,
        KeysView,
        Mapping,
        Optional,
        Tuple,
        Union,
        ValuesView,
    )

__all__ = [
    "BaseDict",
    "BaseProtectedDict",
    "BaseInteractiveDict",
    "BaseMutableDict",
]


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.
VT_co = TypeVar("VT_co", covariant=True)  # Value type covariant containers.


class BaseDict(BaseCollection[KT], SlottedMapping, Generic[KT, VT_co]):
    """
    Base dictionary collection.

    Inherits from:
      - :class:`objetto.bases.BaseCollection`
      - :class:`slotted.SlottedMapping`
      - :class:`typing.Generic`

    Inherited By:
      - :class:`objetto.bases.BaseProtectedDict`
    """

    __slots__ = ()

    def __hash__(self):
        """
        Prevent hashing (not hashable by default).

        :raises TypeError: Not hashable.
        """
        error = "unhashable type: '{}'".format(type(self).__fullname__)
        raise TypeError(error)

    @abstractmethod
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def __reversed__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over reversed keys.

        :return: Reversed keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def __getitem__(self, key):
        # type: (KT) -> VT_co
        """
        Get value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Value.

        :raises KeyError: Key is not present.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self, key, fallback=None):
        # type: (KT, Any) -> Union[VT_co, Any]
        """
        Get value for key, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def iteritems(self):
        # type: () -> Iterator[Tuple[KT, VT_co]]
        """
        Iterate over items.

        :return: Items iterator.
        :rtype: collections.abc.Iterator[tuple[collections.abc.Hashable, Any]]

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def iterkeys(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def itervalues(self):
        # type: () -> Iterator[VT_co]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @final
    def items(self):
        # type: () -> ItemsView[KT, VT]
        """
        Get items.

        :return: Items.
        :rtype: collections.abc.ItemsView
        """
        return collections_abc.ItemsView(self)

    @final
    def keys(self):
        # type: () -> KeysView[KT]
        """
        Get keys.

        :return: Keys.
        :rtype: collections.abc.KeysView
        """
        return collections_abc.KeysView(self)

    @final
    def values(self):
        # type: () -> ValuesView[VT]
        """
        Get values.

        :return: Values.
        :rtype: collections.abc.ValuesView
        """
        return collections_abc.ValuesView(self)


# noinspection PyCallByClass
type.__setattr__(BaseDict, "__hash__", None)  # force non-hashable


# noinspection PyTypeChecker
_BPD = TypeVar("_BPD", bound="BaseProtectedDict")


class BaseProtectedDict(BaseDict[KT, VT], BaseProtectedCollection[KT]):
    """
    Base protected dictionary collection.

    Inherits from:
      - :class:`objetto.bases.BaseDict`
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveDict`
      - :class:`objetto.bases.BaseMutableDict`
      - :class:`objetto.bases.BaseDictStructure`
    """

    __slots__ = ()

    @abstractmethod
    def _discard(self, key):
        # type: (_BPD, KT) -> _BPD
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedDict

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _remove(self, key):
        # type: (_BPD, KT) -> _BPD
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedDict

        :raises KeyError: Key is not present.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _set(self, key, value):
        # type: (_BPD, KT, VT) -> _BPD
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedDict

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def _update(self, __m, **kwargs):
        # type: (_BPD, Mapping[KT, VT], VT) -> _BPD
        pass

    @overload
    @abstractmethod
    def _update(self, __m, **kwargs):
        # type: (_BPD, Iterable[Tuple[KT, VT]], VT) -> _BPD
        pass

    @overload
    @abstractmethod
    def _update(self, **kwargs):
        # type: (_BPD, VT) -> _BPD
        pass

    @abstractmethod
    def _update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedDict

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyTypeChecker
_BID = TypeVar("_BID", bound="BaseInteractiveDict")


# noinspection PyAbstractClass
class BaseInteractiveDict(BaseProtectedDict[KT, VT], BaseInteractiveCollection[KT]):
    """
    Base interactive dictionary collection.

    Inherits from:
      - :class:`objetto.bases.BaseProtectedDict`
      - :class:`objetto.bases.BaseInteractiveCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveDictStructure`
      - :class:`objetto.states.DictState`
    """

    __slots__ = ()

    @final
    def discard(self, key):
        # type: (_BID, KT) -> _BID
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveDict
        """
        return self._discard(key)

    @final
    def remove(self, key):
        # type: (_BID, KT) -> _BID
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveDict

        :raises KeyError: Key is not present.
        """
        return self._remove(key)

    @final
    def set(self, key, value):
        # type: (_BID, KT, VT) -> _BID
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveDict
        """
        return self._set(key, value)

    @overload
    def update(self, __m, **kwargs):
        # type: (_BID, Mapping[KT, VT], VT) -> _BID
        pass

    @overload
    def update(self, __m, **kwargs):
        # type: (_BID, Iterable[Tuple[KT, VT]], VT) -> _BID
        pass

    @overload
    def update(self, **kwargs):
        # type: (_BID, VT) -> _BID
        pass

    @final
    def update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveDict
        """
        return self._update(*args, **kwargs)


class BaseMutableDict(
    SlottedMutableMapping, BaseProtectedDict[KT, VT], BaseMutableCollection[KT]
):
    """
    Base mutable dictionary collection.

    Inherits from:
      - :class:`slotted.SlottedMutableMapping`
      - :class:`objetto.bases.BaseProtectedDict`
      - :class:`objetto.bases.BaseMutableCollection`

    Inherited By:
      - :class:`objetto.bases.BaseMutableDictStructure`
      - :class:`objetto.objects.ProxyDictObject`
    """

    __slots__ = ()

    @final
    def __setitem__(self, key, value):
        # type: (KT, VT) -> None
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.
        """
        self._set(key, value)

    @final
    def __delitem__(self, key):
        # type: (KT) -> None
        """
        Delete key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :raises KeyError: Key is not preset.
        """
        self._remove(key)

    @final
    def clear(self):
        # type: () -> None
        """Clear."""
        self._clear()

    @abstractmethod
    def pop(self, key, fallback=MISSING):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key and remove it, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.

        :raises KeyError: Key is not present and fallback value not provided.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def popitem(self):
        # type: () -> Tuple[KT, VT]
        """
        Get item and discard key.

        :return: Item.
        :rtype: tuple[collections.abc.Hashable, Any]

        :raises KeyError: Dictionary is empty.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def setdefault(self, key, default=None):
        # type: (KT, Optional[VT]) -> Optional[VT]
        """
        Get the value for the specified key, insert key with default if not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param default: Default value.

        :return: Existing or default value.

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @final
    def discard(self, key):
        # type: (KT) -> None
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable
        """
        self._discard(key)

    @final
    def remove(self, key):
        # type: (KT) -> None
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :raises KeyError: Key is not present.
        """
        self._remove(key)

    @final
    def set(self, key, value):
        # type: (KT, VT) -> None
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.
        """
        self._set(key, value)

    @overload
    def update(self, __m, **kwargs):
        # type: (Mapping[KT, VT], VT) -> None
        pass

    @overload
    def update(self, __m, **kwargs):
        # type: (Iterable[Tuple[KT, VT]], VT) -> None
        pass

    @overload
    def update(self, **kwargs):
        # type: (VT) -> None
        pass

    @final
    def update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.
        """
        self._update(*args, **kwargs)



# -*- coding: utf-8 -*-
"""Constants."""

from ._applications import Phase
from ._objects import DELETED
from .utils.lazy_import import (
    IMPORT_PATH_REGEX,
    PARTIAL_IMPORT_PATH_REGEX,
    RELATIVE_IMPORT_PATH_REGEX,
)

__all__ = [
    "DELETED",
    "PRE",
    "POST",
    "PARTIAL_IMPORT_PATH_REGEX",
    "RELATIVE_IMPORT_PATH_REGEX",
    "IMPORT_PATH_REGEX",
]


PRE = Phase.PRE
"""
Phase before the changes are applied.

Inherits from:
  - :class:`objetto.bases.Phase`
"""

POST = Phase.POST
"""
Phase after the changes are applied.

Inherits from:
  - :class:`objetto.bases.Phase`
"""



# -*- coding: utf-8 -*-
"""Object changes."""

from ._changes import (
    Batch,
    DictUpdate,
    ListDelete,
    ListInsert,
    ListMove,
    ListUpdate,
    SetRemove,
    SetUpdate,
    Update,
)

__all__ = [
    "Batch",
    "Update",
    "DictUpdate",
    "ListInsert",
    "ListDelete",
    "ListUpdate",
    "ListMove",
    "SetUpdate",
    "SetRemove",
]



# -*- coding: utf-8 -*-
"""Immutable state types."""

from .bases import BaseState
from .dict import DictState
from .list import ListState
from .set import SetState

__all__ = ["BaseState", "DictState", "ListState", "SetState"]
